/***************************************************************************
 *   Copyright (C) 2009-2016 by Marcel Hasler                              *
 *   mahasler@gmail.com                                                    *
 *                                                                         *
 *   This file is part of KDE CDEmu Manager.                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program. If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "cdemu.h"
#include "devicelistitem.h"
#include "mainwindow.h"
#include "messagebox.h"

#include <KHelpMenu>
#include <KMessageBox>
#include <KStandardAction>

#include <QFileDialog>
#include <QHeaderView>
#include <QPushButton>
#include <QTreeWidget>

static const QString fileTypes("Images (*.mds *.mdx *.b5t *.b6t *.ccd *.sub *.img *.cue *.bin *.toc *.cdi *.cif *.c2d *.iso *.nrg *.udf);;"
                               "Containers (*.dmg *.cdr *.cso *.ecm *.gz *.gbi *.daa *.isz *.xz)");

static const QString historyKey = "history";
static const int maxHistorySize = 10;

static const QString showTrayIconKey = "showTrayIcon";
static const QString lastFilePathKey = "lastFilePath";

MainWindow::MainWindow(QWidget *parent)
    : KMainWindow(parent),
      m_statusLabel(0),
      m_helpMenu(0),
      m_trayIcon(0)
{
    m_ui.setupUi(this);

    // Menus
    m_ui.menuFile->addAction(KStandardAction::quit(qApp, SLOT(quit()), this));

    updateHistory();

    m_helpMenu = new KHelpMenu(this);
    menuBar()->addMenu(m_helpMenu->menu());

    // Tray icon
    QSettings settings;

    bool showTrayIcon = settings.value(showTrayIconKey, true).toBool();
    slotToggleTrayIcon(showTrayIcon);

    m_ui.actionTrayIcon->setChecked(showTrayIcon);
    connect(m_ui.actionTrayIcon, SIGNAL(toggled(bool)), this, SLOT(slotToggleTrayIcon(bool)));

    // Device list
    m_ui.DeviceList->header()->setStretchLastSection(false);

    m_ui.DeviceList->header()->setSectionResizeMode(0, QHeaderView::Fixed);
    m_ui.DeviceList->header()->setSectionResizeMode(1, QHeaderView::Stretch);

    int width = QFontMetrics(font()).width(i18n(m_ui.DeviceList->headerItem()->text(0).toLocal8Bit()) + "xxx");
    m_ui.DeviceList->header()->resizeSection(0, width);

    // Device handling
    connect(m_ui.AddDevice, SIGNAL(clicked()), this, SLOT(slotAddDevice()));
    connect(m_ui.RemoveDevice, SIGNAL(clicked()), this, SLOT(slotRemoveDevice()));

    QObject::connect(CDEmu::instance(), SIGNAL(deviceChanged(unsigned int)), this, SLOT(slotDeviceChanged(unsigned int)));
    QObject::connect(CDEmu::instance(), SIGNAL(daemonChanged(bool)), this, SLOT(slotDaemonChanged(bool)));

    // Status bar
    m_statusLabel = new QLabel(this);
    m_statusLabel->setIndent(10);
    statusBar()->addWidget(m_statusLabel);
    slotDaemonChanged(CDEmu::instance()->isDaemonRunning());

    // Remember window size, etc.
    setAutoSaveSettings();
}


MainWindow::~MainWindow()
{
}


void MainWindow::slotDeviceChanged(unsigned int index)
{
    DeviceListItem *item = dynamic_cast<DeviceListItem*>(m_ui.DeviceList->topLevelItem(index));
    Q_ASSERT(item != 0);

    QString filename = CDEmu::instance()->fileName(index);
    item->setFileName(filename);
}


void MainWindow::slotDaemonChanged(bool running)
{
    Q_ASSERT(m_statusLabel != 0);

    m_ui.DeviceList->setEnabled(running);

    if (running)
    {
        updateDeviceList();
        m_statusLabel->setText(i18n("CDEmu daemon is running."));
    }
    else
        m_statusLabel->setText(i18n("CDEmu daemon not running."));
}


void MainWindow::slotMount(unsigned int index)
{
    QSettings settings;
    QString path = settings.value(lastFilePathKey, QDir::homePath()).toString();

    QString filename = QFileDialog::getOpenFileName(this, i18n("Select an image file"), path, fileTypes);
    if (filename.isEmpty())
        return;

    path = QFileInfo(filename).path();
    settings.setValue(lastFilePathKey, path);

    CDEmu::Result result = CDEmu::instance()->mount(filename, index);

    if (result != CDEmu::Success)
        MessageBox::error(result);
    else
        appendHistory(filename);
}


void MainWindow::slotUnmount(unsigned int index)
{
    CDEmu::Result result = CDEmu::instance()->unmount(index);

    if (result != CDEmu::Success)
        MessageBox::error(result);
}


void MainWindow::slotMountFromHistory()
{
    QAction *action = qobject_cast<QAction*>(sender());
    Q_ASSERT(action);

    int index = CDEmu::instance()->nextFreeDevice();
    if (index < 0)
        MessageBox::error(CDEmu::NoFreeDevice);
    else
    {
        QString filename = action->data().toString();

        CDEmu::Result result = CDEmu::instance()->mount(filename, index);

        if (result != CDEmu::Success)
            MessageBox::error(result);
        else
            appendHistory(filename);
    }
}


void MainWindow::slotClearHistory()
{
    QSettings settings;
    settings.setValue(historyKey, QStringList());

    updateHistory();
}


void MainWindow::slotAddDevice()
{
    CDEmu::Result result = CDEmu::instance()->addDevice();

    if (result != CDEmu::Success)
        MessageBox::error(result);
    else
        updateDeviceList();
}


void MainWindow::slotRemoveDevice()
{
    CDEmu::Result result = CDEmu::instance()->removeDevice();

    if (result != CDEmu::Success)
        MessageBox::error(result);
    else
        updateDeviceList();
}


void MainWindow::slotToggleTrayIcon(bool show)
{
    Q_ASSERT(m_helpMenu != 0);

    if (show && m_trayIcon == 0)
    {
        m_trayIcon = new KStatusNotifierItem(this);
        m_trayIcon->setIconByName("media-optical");
        m_trayIcon->setCategory(KStatusNotifierItem::ApplicationStatus);
        m_trayIcon->setStatus(KStatusNotifierItem::Active);
        m_trayIcon->setToolTip("media-optical", "KDE CDEmu Manager", "");
        m_trayIcon->contextMenu()->addMenu(m_helpMenu->menu());
    }
    else if (!show && m_trayIcon != 0)
    {
        delete m_trayIcon;
        m_trayIcon = 0;
    }

    // Keep running in tray if enabled
    setAttribute(Qt::WA_DeleteOnClose, !show);

    QSettings settings;
    settings.setValue(showTrayIconKey, show);
}


void MainWindow::updateDeviceList()
{
    m_ui.DeviceList->clear();

    int deviceCount = CDEmu::instance()->deviceCount();
    for (int i = 0; i < deviceCount; ++i)
    {
        DeviceListItem *item = new DeviceListItem(i);
        item->setFileName(CDEmu::instance()->fileName(i));

        connect(item, SIGNAL(mountClicked(unsigned int)), this, SLOT(slotMount(unsigned int)));
        connect(item, SIGNAL(unmountClicked(unsigned int)), this, SLOT(slotUnmount(unsigned int)));

        m_ui.DeviceList->addTopLevelItem(item);
        m_ui.DeviceList->setItemWidget(item, 0, new QLabel(QString("  %1").arg(i)));
        m_ui.DeviceList->setItemWidget(item, 1, item->widget());
    }
}


void MainWindow::appendHistory(QString filename)
{
    QSettings settings;
    QStringList history = settings.value(historyKey).toStringList();

    history.removeAll(filename);
    history.prepend(filename);

    settings.setValue(historyKey, history);

    updateHistory();
}


void MainWindow::updateHistory()
{
    QSettings settings;
    QStringList history = settings.value(historyKey).toStringList();

    // Delete invalid entries
    QStringList::iterator it = history.begin();
    while (it != history.end())
    {
        if (!QDir().exists(*it))
            it = history.erase(it);
        else
            ++it;
    }

    // Limit size
    while (history.size() > maxHistorySize)
        history.removeLast();

    settings.setValue(historyKey, history);

    // Rebuild menu
    m_ui.menuHistory->clear();
    for (int i = 0; i < history.size(); ++i)
    {
        QAction *action = m_ui.menuHistory->addAction(history.at(i));
        action->setData(history.at(i));
        connect(action, SIGNAL(triggered(bool)), this, SLOT(slotMountFromHistory()));
    }

    m_ui.menuHistory->addSeparator();

    QAction *action = m_ui.menuHistory->addAction(i18n("Clear history"));
    action->setIcon(QIcon::fromTheme("edit-clear-history"));
    action->setEnabled(!history.isEmpty());
    connect(action, SIGNAL(triggered(bool)), this, SLOT(slotClearHistory()));
}
